# Written by Bram Cohen
# see LICENSE.txt for license information
# Updated and modified for ABC_OKC : Old King Cole

import sys
from os import path, makedirs
from urlparse import urlparse
from time import clock
from threading import Thread, Event
from StringIO import StringIO
from traceback import print_exc

from zurllib import urlopen
from BT1.Choker import Choker
from BT1.Storage import Storage
from BT1.StorageWrapper import StorageWrapper
from BT1.FileSelector import FileSelector
from BT1.Uploader import Upload
from BT1.Downloader import Downloader
from BT1.HTTPDownloader import HTTPDownloader
from BT1.Connecter import Connecter
from RateLimiter import RateLimiter
from BT1.Encrypter import Encoder
from RawServer import RawServer, autodetect_ipv6, autodetect_socket_style
from BT1.Rerequester import Rerequester
from BT1.DHTRerequester import DHTRerequester
from BT1.DownloaderFeedback import DownloaderFeedback
from CurrentRateMeasure import Measure
from BT1.PiecePicker import PiecePicker
from BT1.Statistics import Statistics
from ConfigDir import ConfigDir
from utility import exceptionArgsToString, sysencoding, getABCUtility
from bencode import bencode, bdecode
from parseargs import parseargs
from BTcrypto import CRYPTO_OK
from __init__ import createPeerID


defaults = [
    ('max_uploads', 7,
        "the maximum number of uploads to allow at once."),
    ('floor_uploads', 2,
        "the minimum value for the maximum number of uploads to allow at once."),
    ('keepalive_interval', 120.0,
        'number of seconds to pause between sending keepalives'),
    ('download_slice_size', 2 ** 14,
        "How many bytes to query for per request."),
    ('upload_unit_size', 1460,
    #('upload_unit_size', 1380,
    #('upload_unit_size', 1024,
        "when limiting upload rate, how many bytes to send at a time"),
    ('request_backlog', 10,
        "maximum number of requests to keep in a single pipe at once."),
    ('max_message_length', 2 ** 23,
        "maximum length prefix encoding you'll accept over the wire - larger values get the connection dropped."),
    ('ip', '',
        "ip to report you have to the tracker."),
    ('minport', 10000, 'minimum port to listen on, counts up if unavailable'),
    ('maxport', 60000, 'maximum port to listen on'),
    ('random_port', 1, 'whether to choose randomly inside the port range ' +
        'instead of counting up linearly'),
    ('responsefile', '',
        'file the server response was stored in, alternative to url'),
    ('url', '',
        'url to get file from, alternative to responsefile'),
    ('crypto_allowed', int(CRYPTO_OK),
        'whether to allow the client to accept encrypted connections'),
    ('crypto_only', 0,
        'whether to only create or allow encrypted connections'),
    ('crypto_stealth', 0,
        'whether to prevent all non-encrypted connection attempts; ' +
        'will result in an effectively firewalled state on older trackers'),
    ('selector_enabled', 1,
        'whether to enable the file selector and fast resume function'),
    ('dir_root', 'c:\\',
        'root dir for data cache'),
    ('expire_cache_data', 10,
        'the number of days after which you wish to expire old cache data ' +
        '(0 = disabled)'),
    ('priority', '',
        'a list of file priorities separated by commas, must be one per file, ' +
        '0 = highest, 1 = normal, 2 = lowest, -1 = download disabled'),
    ('saveas', '',
        'local file name to save the file as, null indicates query user'),
    ('timeout', 300.0,
        'time to wait between closing sockets which nothing has been received on'),
    ('timeout_check_interval', 60.0,
        'time to wait between checking if any connections have timed out'),
    ('max_slice_length', 2 ** 17,
        "maximum length slice to send to peers, larger requests are ignored"),
    ('max_rate_period', 10.0,
        "maximum amount of time to guess the current rate estimate represents"),
    ('bind', '', 
        'comma-separated list of ips/hostnames to bind to locally'),
#    ('ipv6_enabled', autodetect_ipv6(),
    ('ipv6_enabled', 0,
         'allow the client to connect to peers via IPv6'),
    ('ipv6_binds_v4', autodetect_socket_style(),
        "set if an IPv6 server socket won't also field IPv4 connections"),
    ('upnp_nat_access', 1,
        'attempt to autoconfigure a UPnP router to forward a server port ' +
        '(0 = disabled, 1 = mode 1 [fast], 2 = mode 2 [slow])'),
    ('upload_rate_fudge', 5.0, 
        'time equivalent of writing to kernel-level TCP buffer, for rate adjustment'),
    #('tcp_ack_fudge', 0.03,
    ('tcp_ack_fudge', 0,
        'how much TCP ACK download overhead to add to upload rate calculations ' +
        '(0 = disabled)'),
    ('display_interval', .5,
        'time between updates of displayed information'),
    ('rerequest_interval', 5 * 60,
        'time to wait between requesting more peers'),
    ('dhtrerequest_interval', 5 * 60,
        'time to wait between requesting more peers in DHT'),
    ('min_peers', 20, 
        'minimum number of peers to not do rerequesting'),
    ('http_timeout', 60, 
        'number of seconds to wait before assuming that an http connection has timed out'),
    ('max_initiate', 40,
        'number of peers at which to stop initiating new connections'),
    ('check_hashes', 1,
        'whether to check hashes on disk'),
    ('max_upload_rate', 0,
        'maximum kB/s to upload at (0 = no limit, -1 = automatic)'),
    ('max_download_rate', 0,
        'maximum kB/s to download at (0 = no limit)'),
    ('alloc_type', 'normal',
        'allocation type (may be normal, background, pre-allocate or sparse)'),
    ('alloc_rate', 2.0,
        'rate (in MiB/s) to allocate space at using background allocation'),
    ('write_buffer_size', 2,
        'the maximum amount of space to use for buffering disk writes ' +
        '(in megabytes, 0 = disabled)'),
    ('write_buffer_expiration', 120,
        'the elapsed time when the write buffer is totally flushed if no partial flush has occured before' +
        '(in seconds)'),
    ('read_buffer_expiration', 120,
        'the elapsed time when the read buffer is expired when it is not accessed' +
        '(in seconds)'),
    ('read_buffer_pool_expiration', 120,
        'the elapsed time when buffers are deleted from the read buffer pool when they are not re-used' +
        '(in seconds)'),
    ('breakup_seed_bitfield', 1,
        'sends an incomplete bitfield and then fills with have messages, '
        'in order to get around stupid ISP manipulation'),
    ('snub_time', 30.0,
        "seconds to wait for data to come in over a connection before assuming it's semi-permanently choked"),
    ('spew', 1,
        "whether to display diagnostic info to stdout"),
    ('rarest_first_cutoff', 2,
        "number of downloads at which to switch from random to rarest first"),
    ('rarest_first_priority_cutoff', 5,
        'the number of peers which need to have a piece before other partials take priority over rarest first'),
    ('min_uploads', 4,
        "the number of uploads to fill out to with extra optimistic unchokes"),
    ('max_files_open', 50,
        'the maximum number of files to keep open at a time, 0 means no limit'),
    ('round_robin_period', 30,
        "the number of seconds between the client's switching upload targets"),
    ('super_seeder', 0,
        "whether to use special upload-efficiency-maximizing routines (only for dedicated seeds)"),
    ('security', 1,
        "whether to enable extra security features intended to prevent abuse"),
    ('max_connections', 0,
        "the absolute maximum number of peers to connect with (0 = no limit)"),
    ('auto_kick', 1,
        "whether to allow the client to automatically kick/ban peers that send bad data"),
    ('double_check', 1,
        "whether to double-check data being written to the disk for errors (may increase CPU load)"),
    ('triple_check', 0,
        "whether to thoroughly check data being written to the disk (may slow disk access)"),
    ('lock_files', 1,
        "whether to lock files the client is working with"),
    ('lock_while_reading', 0,
        "whether to lock access to files being read"),
    #('auto_flush', 0,
    #    "minutes between automatic flushes to disk (0 = disabled)"),
    ('dedicated_seed_id', '',
        "code to send to tracker identifying as a dedicated seed"),
    ('dhtport', 56666, 'DHT port'),
    ('default_tracker', '', 'default tracker url'),
    ('magnet', 0, 'if torrent is created from magnet and still has incomplete metainfo'),
    ('metadata_timeout', 60, 'timeout to receive torrent metadata'),
    ('ut_pex', 0, 'whether to support PEX'),
    ('ut_pex_sending_rate', 60, 'sending rate for PEX messages'),
    ('read_buffer_memory_saving', 0, 'read buffer memory optimization method'),
    ('multihome_ipv4', '', 'extra ipv4 address to add to announce for multihomed host'),
    ('multihome_ipv6', '', 'extra ipv6 address to add to announce for multihomed host')
    ]


def parse_params(params, presets = {}):
    if len(params) == 0:
        return None
    config, args = parseargs(params, defaults, 0, 1, presets = presets)
    if args:
        if config['responsefile'] or config['url']:
            raise ValueError,getABCUtility().lang.get('argorparam')
        if path.isfile(args[0]):
            config['responsefile'] = args[0]
        else:
            try:
                urlparse(args[0])
            except:
                raise ValueError, getABCUtility().lang.get('badfile')
            config['url'] = args[0]
    elif (config['responsefile'] == '') == (config['url'] == ''):
        raise ValueError, getABCUtility().lang.get('responseorurl')
    return config


def get_response(file, errorfunc):
    try:
        h = open(file, 'rb')
        response = h.read()
    except IOError, e:
        try:    
            h.close()
        except:
            pass
        errorfunc(getABCUtility().lang.get('cantgetresponse') + ' - ' + exceptionArgsToString(e))
        return None
    h.close()
    try:
        response = bdecode(response, sloppy = 1)
    except ValueError, e:
        errorfunc(getABCUtility().lang.get('badresponse') + " - " + exceptionArgsToString(e))
        return None
    return response


class BT1Download:
    def __init__(self, statusfunc, finfunc, errorfunc, excfunc, doneflag,
                 config, response, infohash, id, rawserver, port, dht,
                 appdataobj = None, extann = None, trackerokfunc = None,
                 endallocfunc = None, rejectexceptions = [], filesnamefunc = None,
                 resetcompletefunc = None, metadatacompletefunc = None, reseed = False,
                 readbufferallowedfunc = None, checkfreediskspacefunc = None):
        self.statusfunc = statusfunc
        self.finfunc = finfunc
        self.errorfunc = errorfunc
        self.excfunc = excfunc
        self.doneflag = doneflag
        self.config = config
        self.response = response
        self.infohash = infohash
        self.myid = id
        self.rawserver = rawserver
        self.port = port
        self.dht = dht
        self.reseed = reseed
        self.info = self.response['info']
        self.piecelength = self.info['piece length']
        if self.config['magnet']:
            self.metadata = []
        else:
            self.metadata = bencode(self.info)
        self.pieces = [self.info['pieces'][x:x + 20]
                       for x in xrange(0, len(self.info['pieces']), 20)]
        self.private = self.info.get('private')
        if self.private is None:
            self.private = 0
        self.len_pieces = len(self.pieces)
        self.unpauseflag = Event()
        self.unpauseflag.set()
        self.storage = None
        self.storagewrapper = None
        self.fileselector = None
        self.downloader = None
        self.connecter = None
        self.encoder = None
        self.httpdownloader = None
        self.statistics = None
        self.downloaderfeedback = None
        self.ratelimiter = None
        self.super_seeding_active = False
        self.filedatflag = Event()
        self.spewflag = Event()
        self.superseedflag = Event()
        self.whenpaused = None
        self.finflag = Event()
        self.rerequest = None
        self.dhtrerequest = None
        self.extann = extann
        self.tcp_ack_fudge = config['tcp_ack_fudge']
        self.trackerokfunc = trackerokfunc
        self.endallocfunc = endallocfunc
        self.rejectexceptions = rejectexceptions
        self.filesnamefunc = filesnamefunc
        self.resetcompletefunc = resetcompletefunc
        self.metadatacompletefunc = metadatacompletefunc
        self.readbufferallowedfunc = readbufferallowedfunc
        self.checkfreediskspacefunc = checkfreediskspacefunc
        self.updatestorageflag = None
        # To deal with the case when BT has downloaded all files of a torrent even
        # those marked "Never download" (often very small files) and wants to finish
        # In this case we force it to keep on running
        self.forcerunning = False
        self.datalength = 0

        self.selector_enabled = config['selector_enabled']
        if appdataobj:
            self.appdataobj = appdataobj
        elif self.selector_enabled:
            self.appdataobj = ConfigDir(config['dir_root'])
            self.appdataobj.deleteOldCacheData(config['expire_cache_data'])

        self.failed = False
        self.storageinit = False
        self.waiting = True
        self.started = False

        self.picker = PiecePicker(self.len_pieces, config)
        self.choker = Choker(self, config, rawserver.add_task,
                             self.picker, self.piecelength,
                             self.readbufferallowedfunc,
                             self.finflag.isSet)

    def updateData(self, src, dest, private):
        self.config['responsefile'] = src
        self.config['saveas'] = dest
        self.response = get_response(src, self.errorfunc)
        self.info = self.response['info']
        self.piecelength = self.info['piece length']
        self.private = private
        self.pieces = [self.info['pieces'][x:x + 20]
                       for x in xrange(0, len(self.info['pieces']), 20)]
        self.len_pieces = len(self.pieces)
        self.waiting = True
        self.choker.updateData(self.piecelength)
        self.picker.updateData(self.len_pieces)
        self.downloader.updateData(self.len_pieces)
        self.httpdownloader.updateData()
        self.connecter.updateData(self.len_pieces, private)
        if self.dht and private:
            self.dhtrerequest_stopped()

    def updateStorage(self, filefunc, sem, unblockwaiting, unblockwaitfunc, updatetoragecompletefunc):
        self.rawserver.add_task(self._updateStorage, 0, [filefunc, sem, unblockwaiting, unblockwaitfunc, updatetoragecompletefunc])

    def _updateStorage(self, filefunc, sem, unblockwaiting, unblockwaitfunc, updatetoragecompletefunc):
        # Thread to update storage while raw server keeps on running
        self.updatestorageflag = Event()
        thread = Thread(target = self.__updateStorage, args = [filefunc, sem, unblockwaiting, unblockwaitfunc, updatetoragecompletefunc])
        thread.daemon = False
        thread.start()

    def __updateStorage(self, filefunc, sem, unblockwaiting, unblockwaitfunc, updatetoragecompletefunc):
        # Update storage with new data
        if not self.saveAs(filefunc):
            self.updatestorageflag.set()
            return
        self.config['priority'] = ','.join(len(self.files) * ['1'])
        self.priority = len(self.files) * [1]
        if not self.updateFiles(sem, unblockwaiting, unblockwaitfunc):
            self.updatestorageflag.set()
            return
        self.rawserver.add_task(self.___updateStorage)
        updatetoragecompletefunc()
        self.encoder.onMetadataOK()
        self.updatestorageflag.set()

    def ___updateStorage(self):
        self.statistics.updateData()
        if self.info.has_key('files'):
            self.statistics.set_dirstats(self.files, self.piecelength)
        for i in xrange(self.len_pieces):
            if self.storagewrapper.do_I_have(i):
                self.picker.complete(i)
        if self.selector_enabled:
            self.fileselector.tie_in(self.picker, self._cancelfunc,
                    self._reqmorefunc, self.rerequest_ondownloadmore)
            if self.priority:
                self.fileselector.set_priorities_now(self.priority)
            # Erase old data once you've started modifying it
            self.appdataobj.deleteTorrentData(self.infohash)

    def getReadBufferSize(self):
        if not self.connecter:
            return 0
        return len([c for c in self.connecter.connections.values() if c.upload.piecebuf is not None]) * self.piecelength

    def checkSaveLocation(self, loc):
        if self.info.has_key('length'):
            return path.exists(loc)
        for x in self.info['files']:
            if path.exists(path.join(loc, x['path'][0])):
                return True
        return False

    def saveAs(self, filefunc, pathfunc = None):
        try:
            def make(f, forcedir = False):
                if not forcedir:
                    f = path.split(f)[0]
                if f != '' and not path.exists(f):
                    makedirs(f)

            if self.info.has_key('name.utf-8'):
                namekey = 'name.utf-8'
            else:
                namekey = 'name'
            try:
                file_name = self.info[namekey].decode('utf_8')
            except:
                file_name = self.info[namekey].decode(sysencoding)
            if self.info.has_key('length'):
                file_length = self.info['length']
                file = filefunc(file_name, file_length,
                                self.config['saveas'], False)
                if file is None:
                    return None
                make(file)
                files = [(file, file_length)]
            else:
                file_length = 0L
                for x in self.info['files']:
                    file_length += x['length']
                file = filefunc(file_name, file_length,
                                self.config['saveas'], True)
                if file is None:
                    return None

                ## if this path exists, and no files from the info dict exist, we assume it's a new download and 
                ## the user wants to create a new directory with the default name
                #existing = 0
                #if path.exists(file):
                #    if not path.isdir(file):
                #        self.errorfunc(file + 'is not a dir')
                #        return None
                #    if len(listdir(file)) > 0:  # if it's not empty
                #        for x in self.info['files']:
                #            if path.exists(path.join(file, x['path'][0])):
                #                existing = 1
                #        if not existing:
                #            file = path.join(file, self.info['name'])
                #            if path.exists(file) and not path.isdir(file):
                #                if file[-8:] == '.torrent':
                #                    file = file[:-8]
                #                if path.exists(file) and not path.isdir(file):
                #                    self.errorfunc("Can't create dir - " + self.info['name'])
                #                    return None
                # ===== All commented above replaced by following to have every torrents saved into their own directory
                if path.exists(file) and not path.isdir(file):
                    self.errorfunc(file + getABCUtility().lang.get('notadir2'))
                    return None
                #else:
                #    file = path.join(file, self.info['name'])
                #    if path.exists(file) and not path.isdir(file):
                #        if file[-8:] == '.torrent':
                #            file = file[:-8]
                #        if path.exists(file) and not path.isdir(file):
                #            self.errorfunc("Can't create dir - " + self.info['name'])
                #            return None
                # ===== End of replace

                make(file, True)

                # alert the UI to any possible change in path
                if pathfunc != None:
                    pathfunc(file)

                filesname = self.filesnamefunc(self.info['files'])
                files = []
                i = 0
                for f in self.info['files']:
                    filepath = path.join(file, filesname[i])
                    files.append((filepath, f['length']))
                    make(filepath)
                    i += 1

        except OSError, e:
            self.errorfunc(getABCUtility().lang.get('cantallocatedir') + " - " + exceptionArgsToString(e))
            return None

        self.filename = file
        self.files = files
        self.datalength = file_length

        return file
    
    def getFilename(self):
        return self.filename

    def setPriorities(self, priority):
        self.rawserver.add_task(self._setPriorities, 0, [priority])

    def _setPriorities(self, priority):
        if self.config['magnet']:
            return
        self.priority = priority[:]
        if self.selector_enabled and self.fileselector:
            self.fileselector.set_priorities(self.priority)

    def _finished(self, initphase = True):
        if -1 in self.priority:
            # If all pieces have been downloaded but some files are marked as
            # Never download, don't consider the torrent as complete
            self.forcerunning = True
            return
        self.finflag.set()
        try:
            self.storage.set_readonly()
        except (IOError, OSError), e:
            self.errorfunc(getABCUtility().lang.get('cantsetreadonly') + ' - ' + exceptionArgsToString(e))
        if self.superseedflag.isSet():
            self._set_super_seed()
        self.choker.set_round_robin_period(
            max( self.config['round_robin_period'],
                 self.config['round_robin_period'] *
                                     self.piecelength / 200000 ) )
        self.rerequest_complete()
        self.finfunc(initphase)

    def _data_flunked(self, amount, index):
        if not self.doneflag.isSet():
            self.errorfunc(getABCUtility().lang.get('piecefailedhashcheck') % index)

    def _failed(self, reason):
        self.failed = True
        self.doneflag.set()
        if reason is not None:
            self.errorfunc(reason)

    def initFiles(self, sem, unblockwaiting, unblockwaitfunc, statusfunc = None):
        if self.doneflag.isSet():
            return None
        if not statusfunc:
            statusfunc = self.statusfunc

        #disabled_files = None
        if self.selector_enabled:
            self.priority = self.config['priority']
            if self.priority:
                try:
                    self.priority = self.priority.split(',')
                    assert len(self.priority) == len(self.files)
                    self.priority = [int(p) for p in self.priority]
                    for p in self.priority:
                        assert p >= -1
                        assert p <= 2
                except:
                    self.errorfunc(getABCUtility().lang.get('badpriority'))
                    self.priority = []
            else:
                self.priority = []

            if self.reseed:
                data = {'resume data': {'pieces': 1}}
            else:
                data = self.appdataobj.getTorrentData(self.infohash)

            #try:
            #    d = data['resume data']['priority']
            #    assert len(d) == numfiles
            #    disabled_files = [x == -1 for x in d]
            #except:
            #    try:
            #        disabled_files = [x == -1 for x in self.priority]
            #    except:
            #        pass

        self.unpauseflag.wait()

        sem.acquire()
        unblockwaitfunc()
        unblockwaiting.wait()

        if self.doneflag.isSet():
            sem.release()
            return None

        if not self.checkfreediskspacefunc(onstart = True):
            sem.release()
            return None

        try:
            try:
                self.storage = Storage(self.files, self.piecelength,
                                       self.doneflag, self.config,
                                       self.len_pieces, self.reseed,
                                       self.rawserver.add_task)
                                       #self.doneflag, self.config, disabled_files)
            except IOError, e:
                self._failed(getABCUtility().lang.get('cantaccessfile') + ' - ' + exceptionArgsToString(e))
            if self.doneflag.isSet():
                sem.release()
                return None
            self.storagewrapper = StorageWrapper(self.storage, self.config['download_slice_size'],
                                                 self.pieces, self.piecelength, self._finished,
                                                 self._failed, statusfunc, self.doneflag, self.unpauseflag,
                                                 self.config['check_hashes'], self._data_flunked,
                                                 self.rawserver.add_task, self.config, self.endallocfunc,
                                                 self.checkfreediskspacefunc)
        except ValueError, e:
            self._failed(getABCUtility().lang.get('baddata') + ' - ' + exceptionArgsToString(e))
        except IOError, e:
            self._failed(getABCUtility().lang.get('ioerror') + ' - ' + exceptionArgsToString(e))
        if self.doneflag.isSet():
            sem.release()
            return None

        if self.selector_enabled:
            self.fileselector = FileSelector(self.files, self.piecelength,
                                             self.appdataobj.getPieceDir(self.infohash),
                                             self.storage, self.storagewrapper,
                                             self.rawserver.add_task,
                                             self._failed)
            if self.priority:
                self.fileselector.init_priority(self.priority)
            if data:
                data = data.get('resume data')
                if data:
                    self.fileselector.unpickle(data)

        if self.doneflag.isSet():
            sem.release()
            return None

        self.resetcompletefunc()

        self.storageinit = True

        try:
            swrapper = self.storagewrapper.old_style_init()
        except IOError, e:
            self._failed(getABCUtility().lang.get('ioerror') + ' - ' + exceptionArgsToString(e))
        except MemoryError, e:
            self._failed(getABCUtility().lang.get('memoryerror') + ' - ' + exceptionArgsToString(e))
        except SystemError, e:
            self._failed(getABCUtility().lang.get('systemerror') + ' - ' + exceptionArgsToString(e))
        except:
            data = StringIO()
            print_exc(file = data)
            self._failed(data.getvalue().decode(sysencoding))

        if self.doneflag.isSet():
            sem.release()
            return None

        sem.release()
        return swrapper

    def updateFiles(self, sem, unblockwaiting, unblockwaitfunc):
        if self.doneflag.isSet():
            return None

        self.unpauseflag.wait()

        sem.acquire()
        unblockwaitfunc()
        unblockwaiting.wait()

        if self.doneflag.isSet():
            sem.release()
            return None

        if not self.checkfreediskspacefunc(onstart = True):
            sem.release()
            return None

        self.storage.updateData(self.files, self.piecelength, self.len_pieces)
        self.storagewrapper.updateData(self.pieces, self.piecelength)
        if self.selector_enabled:
            self.fileselector.updateData(self.files, self.piecelength)

        self.resetcompletefunc()

        self.storageinit = True

        try:
            swrapper = self.storagewrapper.old_style_init(reset = True)
        except IOError, e:
            self._failed(getABCUtility().lang.get('ioerror') + ' - ' + exceptionArgsToString(e))
        except MemoryError, e:
            self._failed(getABCUtility().lang.get('memoryerror') + ' - ' + exceptionArgsToString(e))
        except SystemError, e:
            self._failed(getABCUtility().lang.get('systemerror') + ' - ' + exceptionArgsToString(e))
        except:
            data = StringIO()
            print_exc(file = data)
            self._failed(data.getvalue().decode(sysencoding))

        if self.doneflag.isSet():
            sem.release()
            return None

        sem.release()
        return swrapper

    #def getCachedTorrentData(self):
    #    return self.appdataobj.getTorrentData(self.infohash)

    def _make_upload(self, connection, ratelimiter, readbufferallowedfunc):
        return Upload(connection, ratelimiter,
                      self.choker, self.storagewrapper, self.picker,
                      readbufferallowedfunc,
                      self.config, self.rawserver.add_task)

    def _kick_peer(self, connection):
        self.rawserver.add_task(connection.close)

    def _ban_peer(self, ip):
        self.encoder_ban(ip)

    def _received_raw_data(self, x):
        if self.tcp_ack_fudge:
            x = int(x*self.tcp_ack_fudge)
            self.ratelimiter.adjust_sent(x)

    def _received_data(self, x):
        self.downmeasure.update_rate(x)

    def _received_http_data(self, x):
        self.downmeasure.update_rate(x)
        self.downloader.external_data_received(x)

    def _cancelfunc(self, pieces):
        self.downloader.cancel_piece_download(pieces)
        self.httpdownloader.cancel_piece_download(pieces)

    def _reqmorefunc(self, pieces):
        self.downloader.requeue_piece_download(pieces)

    def startEngine(self, ratelimiter = None, statusfunc = None):
        if self.doneflag.isSet():
            return False
        if not statusfunc:
            statusfunc = self.statusfunc

        if not CRYPTO_OK:
            if self.config['crypto_allowed']:
                self.errorfunc(getABCUtility().lang.get('cryptonotinstalled'))
            self.config['crypto_allowed'] = 0
            self.config['crypto_only'] = 0
            self.config['crypto_stealth'] = 0

        for i in xrange(self.len_pieces):
            if self.storagewrapper.do_I_have(i):
                self.picker.complete(i)

        self.downmeasure = Measure(self.config['max_rate_period'])

        if ratelimiter:
            self.ratelimiter = ratelimiter
        else:
            self.ratelimiter = RateLimiter(self.rawserver.add_task,
                                           self.config['upload_unit_size'],
                                           self.setConns)
            self.ratelimiter.set_upload_rate(self.config['max_upload_rate'])

        self.downloader = Downloader(self.storagewrapper, self.picker, self.len_pieces,
                                     self._received_data, self._kick_peer, self._ban_peer,
                                     self.config)
        self.downloader.set_download_rate(self.config['max_download_rate'])
        self.connecter = Connecter(self._make_upload, self.downloader,
                                   self.choker, self.len_pieces, self.ratelimiter,
                                   self.dht, self.private, self.port,
                                   self.metadata, self.metadatacompletefunc,
                                   self.readbufferallowedfunc,
                                   self.rawserver.add_task, self.config)
        self.encoder = Encoder(self.connecter, self.rawserver, self.myid,
                               self.dht, self.rawserver.add_task, self.infohash,
                               self._received_raw_data, self.config)
        self.encoder_ban = self.encoder.ban

        self.httpdownloader = HTTPDownloader(self.storagewrapper, self.picker, self.rawserver,
                                             self.finflag, self.errorfunc, self.downloader,
                                             self.config['max_rate_period'], self.infohash,
                                             self._received_http_data, self.connecter.got_piece)
        if self.response.has_key('httpseeds') and not self.finflag.isSet():
            for u in self.response['httpseeds']:
                self.httpdownloader.make_download(u)

        if not self.config['magnet'] and self.selector_enabled:
            self.fileselector.tie_in(self.picker, self._cancelfunc,
                    self._reqmorefunc, self.rerequest_ondownloadmore)
            if self.priority:
                self.fileselector.set_priorities_now(self.priority)
            # Erase old data once you've started modifying it
            self.appdataobj.deleteTorrentData(self.infohash)

        if self.config['super_seeder']:
            self.set_super_seed()

        self.started = True
        return True

    def rerequest_complete(self):
        if self.rerequest:
            self.rerequest.announce(1)

    def rerequest_stopped(self):
        if self.rerequest:
            self.rerequest.announce(2)

    def rerequest_lastfailed(self):
        if self.rerequest:
            return self.rerequest.last_failed
        return False

    def rerequest_ondownloadmore(self):
        if self.rerequest:
            self.rerequest.hit()

    def startRerequester(self, seededfunc = None):
        if self.response.has_key('announce-list'):
            trackerlist = self.response['announce-list']
            for tier in trackerlist:
                for t in xrange(len(tier)):
                    try:
                        tier[t] = tier[t].decode('utf_8')
                    except:
                        tier[t] = tier[t].decode(sysencoding)
        else:
            announce = self.response.get('announce')
            if announce is None:
                trackerlist = []
            else:
                try:
                    tracker = announce.decode('utf_8')
                except:
                    tracker = announce.decode(sysencoding)
                trackerlist = [[tracker]]

        self.rerequest = Rerequester(self.port, self.myid, self.infohash,
            trackerlist, self.config, self.rawserver.add_task,
            self.errorfunc, self.excfunc,
            self.encoder.start_connections,
            self.connecter.how_many_connections,
            self.storagewrapper.get_amount_left,
            self.ratelimiter.measure.get_total, self.downmeasure.get_total,
            self.ratelimiter.measure.get_rate, self.downmeasure.get_rate,
            self.doneflag, self.unpauseflag, seededfunc,
            special = self.extann, trackerokfunc = self.trackerokfunc,
            rejectexceptions = self.rejectexceptions)

        self.rerequest.start()

    def dhtrerequest_stopped(self):
        if self.dhtrerequest:
            self.dhtrerequest.stopped = True

    def startDHTRerequester(self):
        if self.config['ipv6_enabled']:
            alltablesempty = True
            if self.dht.table6.findNodes(self.infohash):
                alltablesempty = False
            else:
                self.dht.sched(self.dht.findCloseNodes, 5, [[self.dht.table6]])
            if self.dht.table.findNodes(self.infohash):
                alltablesempty = False
            else:
                self.dht.sched(self.dht.findCloseNodes, 5, [[self.dht.table]])
            if alltablesempty:
                self.rawserver.add_task(self._startDHTRerequester, 20)
            else:
                self._startDHTRerequester()
        else:
            if self.dht.table.findNodes(self.infohash):
                self._startDHTRerequester()
            else:
                # Local table is empty
                self.dht.sched(self.dht.findCloseNodes, 5, [[self.dht.table]])
                self.rawserver.add_task(self._startDHTRerequester, 20)

    def _startDHTRerequester(self):
        if self.response.has_key('nodes'):
            for ip, port in self.response['nodes']:
                resolvedips = getABCUtility().getIpsByName(ip)
                for ri in resolvedips:
                    self.dht.sched(self.dht.addContact, 0, [ri, port])
        self.dhtrerequest = DHTRerequester(self.infohash, self.config, self.dht,
                                           self.rawserver.add_task, self.errorfunc,
                                           self.encoder.start_connections,
                                           self.connecter.how_many_connections,
                                           self.unpauseflag, self.port)
        self.dhtrerequest.start()

    def _init_stats(self):
        self.statistics = Statistics(self.downmeasure, self.connecter,
                                     self.httpdownloader, self.ratelimiter,
                                     self.rerequest_lastfailed, self.filedatflag,
                                     self.config)
        if self.info.has_key('files'):
            self.statistics.set_dirstats(self.files, self.piecelength)
        if self.config['spew']:
            self.spewflag.set()

    def autoStats(self, displayfunc = None):
        if not displayfunc:
            displayfunc = self.statusfunc

        self._init_stats()
        self.downloaderfeedback = DownloaderFeedback(self.storagewrapper, self.choker,
                                                     self.httpdownloader,
                                                     self.rawserver.add_task,
                                                     self.ratelimiter.measure.get_rate,
                                                     self.downmeasure.get_rate,
                                                     self.connecter.pexupmeasure.get_rate,
                                                     self.getReadBufferSize,
                                                     self.storagewrapper.get_stats,
                                                     self.datalength, self.finflag, self.spewflag,
                                                     self.statistics, self.config, displayfunc,
                                                     self.config['display_interval'])

    #def startStats(self):
    #    self._init_stats()
    #    d = DownloaderFeedback(self.storagewrapper, self.choker, self.httpdownloader, self.rawserver.add_task,
    #        self.ratelimiter.measure.get_rate, self.downmeasure.get_rate,
    #        self.storagewrapper.get_stats,
    #        self.datalength, self.finflag, self.spewflag, self.statistics)
    #    return d.gather

    def getPortHandler(self):
        return self.encoder

    def shutdown(self, torrentdata = {}, exception = False):
        if exception:
            self.failed = True
            self.doneflag.set()
        if self.updatestorageflag:
            self.updatestorageflag.wait()
        if self.storageinit or self.started:
            if not exception:
                self.storagewrapper.sync()
            self.storage.close()
        if self.rerequest:
            self.rerequest_stopped()
        if self.dhtrerequest:
            self.dhtrerequest_stopped()
        if self.encoder:
            self.encoder.shutdown()
        if self.connecter:
            self.connecter.shutdown()
        if self.httpdownloader:
            self.httpdownloader.shutdown()
        if self.downloader :
            self.downloader.shutdown()
        if self.fileselector and self.started:
            if not self.failed:
                #if self.info.has_key('files'):
                #    self.fileselector.finish()
                torrentdata['resume data'] = self.fileselector.pickle()
            try:
                self.appdataobj.writeTorrentData(self.infohash,torrentdata)
            except:
                self.appdataobj.deleteTorrentData(self.infohash) # clear it
        self.rawserver = None

    def setUploadRate(self, rate):
        try:
            def s(self = self, rate = rate):
                self.config['max_upload_rate'] = rate
                self.ratelimiter.set_upload_rate(rate)
            self.rawserver.add_task(s)
        except AttributeError:
            pass

    def setConns(self, conns, conns2 = None):
        if not conns2:
            conns2 = conns
        try:
            def s(self = self, conns = conns, conns2 = conns2):
                self.config['min_uploads'] = conns
                self.config['max_uploads'] = conns2
                if (conns > 30):
                    self.config['max_initiate'] = conns + 10
            self.rawserver.add_task(s)
        except AttributeError:
            pass
        
    def setDownloadRate(self, rate):
        try:
            def s(self = self, rate = rate):
                self.config['max_download_rate'] = rate
                self.downloader.set_download_rate(rate)
            self.rawserver.add_task(s)
        except AttributeError:
            pass

    def startConnection(self, ip, port, id):
        self.encoder._start_connection((ip, port), id)
      
    def _startConnection(self, ipandport, id):
        self.encoder._start_connection(ipandport, id)
        
    def setInitiate(self, initiate):
        try:
            def s(self = self, initiate = initiate):
                self.config['max_initiate'] = initiate
            self.rawserver.add_task(s)
        except AttributeError:
            pass

    def reannounce(self, tracker = None, special = None):
        try:
            def r(self = self, tracker = tracker, special = special):
                self.rerequest.announce(tracker = tracker, specialurl = special)
            self.rawserver.add_task(r)
            self.rawserver.add_task(self.dhtrerequest.rerequest, 0, [True])
        except AttributeError:
            pass

    def getResponse(self):
        try:
            return self.response
        except:
            return None

    def Pause(self):
        if not self.storagewrapper:
            return False
        self.unpauseflag.clear()
        self.rawserver.add_task_prio(self.onPause)
        return True

    def onPause(self):
        self.whenpaused = clock()
        if not self.downloader:
            return
        self.downloader.pause(True)
        self.encoder.pause(True)
        self.connecter.pause(True)
        self.choker.pause(True)
    
    def Unpause(self):
        self.unpauseflag.set()
        self.rawserver.add_task(self.onUnpause)

    def onUnpause(self):
        if not self.downloader:
            return
        self.downloader.pause(False)
        self.encoder.pause(False)
        self.connecter.pause(False)
        self.choker.pause(False)
        #if self.rerequest and self.whenpaused and clock()-self.whenpaused > 60:
        #    self.rerequest.announce(3)      # rerequest automatically if paused for >60 seconds
        if self.rerequest and self.rerequest.stopped:
            # Restarts the rerequester so that the rerequester min interval cycle is re-synchronized on the resume
            self.rerequest.start(3)
        if self.dhtrerequest and self.dhtrerequest.stopped:
            self.dhtrerequest.start()

    def set_super_seed(self):
        try:
            self.superseedflag.set()
            def s(self = self):
                if self.finflag.isSet():
                    self._set_super_seed()
            self.rawserver.add_task(s)
        except AttributeError:
            pass

    def _set_super_seed(self):
        if not self.super_seeding_active:
            self.super_seeding_active = True
            self.errorfunc(getABCUtility().lang.get('superseedactive'))
            def s(self = self):
                self.downloader.set_super_seed()
                self.choker.set_super_seed()
            self.rawserver.add_task(s)
            if self.finflag.isSet():        # mode started when already finished
                # so after kicking everyone off, reannounce
                self.rawserver.add_task(self.rerequest.announce, 0, [3])

    def am_I_finished(self):
        return self.finflag.isSet()

    def addPeers(self, peers):
        if self.encoder:
            self.rawserver.add_task(self.encoder.start_connections, 0, [peers])

    def setPex(self, flag):
        if self.connecter:
            self.rawserver.add_task(self.connecter.setPex, 0, [flag])

    def setWriteBufferSize(self, writebuffersize):
        if self.storagewrapper:
            self.rawserver.add_task(self.storagewrapper.setWriteBufferSize, 0, [writebuffersize])

    def setReadBufferMemorySaving(self, mode):
        self.rawserver.add_task(self._setReadBufferMemorySaving, 0, [mode])

    def _setReadBufferMemorySaving(self, mode):
        self.config['read_buffer_memory_saving'] = mode
